import pandas as pd
from pymongo.collection import Collection


def df_2_db(df:pd.DataFrame, collection:Collection) -> None:  # *
    """Inserts dataframe into mongo Collection.

    Args:
        df (pd.DataFrame): Data to insert into collection.
        collection (Collection): Collection object to insert data into.
    """
    data_dict = df.to_dict('records')
    collection.insert_many(data_dict)
    
def db_2_df(collection:Collection) -> pd.DataFrame:  # *
    """Extracts all data from collection and returns as single pandas DataFrame.

    Args:
        collection (Collection): Collection to grab all data from.

    Returns:
        pd.DataFrame: Data in collection.
    """
    cursor = collection.find({}, {'_id': 0})
    return pd.DataFrame(list(cursor))



def branches(data:dict):
    vals = []
    keys = []
    def __base_list(dd, key:str):
        if isinstance(dd, dict):
            [list(__base_list(dd[d], f'{key}_{d}')) for d in list(dd.keys())]
        else:
            vals.append(dd)
            keys.append(key)
            yield dd
    list(__base_list(data, ''))
    
    out = {}
    for k, v in zip(keys, vals):
        out[k] = v
    return out



import unittest
import numpy as np

from stse import bytes_vect


class TestBytes(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        return super().setUpClass()
    
    def test_bit_vect(self):
        # Test non-iterable single index value
        np.testing.assert_array_equal(
            bytes_vect.bit_vect(8, 1), 
            [0, 1, 0, 0, 0, 0, 0, 0]
        )
        
        # Test iterable multi-index value
        np.testing.assert_array_equal(
            bytes_vect.bit_vect(8, [1, 2, 3, 6]), 
            [0, 1, 1, 1, 0, 0, 1, 0]
        )
        
    def test_remove_hot_overlap(self):
        # Test w/ lists
        np.testing.assert_array_equal(
            bytes_vect.remove_hot_overlap([0, 1, 1, 0, 0], [1, 1, 0, 0, 0]),
            [0, 0, 1, 0, 0]
        )
        
        # Test w/ arrays
        np.testing.assert_array_equal(
            bytes_vect.remove_hot_overlap(np.array([0, 1, 1, 0, 0]), np.array([1, 1, 0, 0, 0])),
            [0, 0, 1, 0, 0]
        )


if __name__ == '__main__':
    unittest.main()



import unittest
import pandas as pd
import mongomock

from stse import database as db


class TestDatabase(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        cls.df = pd.DataFrame({
            'col1': [
                1,
                2,
                3,
                4
            ],
            'col2': [
                'a',
                'b',
                'c',
                'd'
            ]
        })

        client = mongomock.MongoClient()
        db = client['test_db']
        cls.coll = db['coll']
        
        return super().setUpClass()

    def test_df_2_db_and_db_2_df(self):
        db.df_2_db(self.df, self.coll)
        out = db.db_2_df(self.coll)
        
        self.assertEqual(
            True,
            self.df.equals(out)
        )


if __name__ == '__main__':
    unittest.main()



import numpy as np


"""
    File name: bytes.py
    Author: Jacob Gerlach
    Description: Assortment of basic byte vector operations.
    Notes:
        * = Function has associated unit test.
"""

def bit_vect(length, indices):  # *
    """Generates a bit vector, hot at each index in list.

    :param length: Length of vector to generate
    :type length: int
    :param indices: Collection of indices to set to 1, all other set to 0
    :type indices: iterable[int]
    :return: one-hot bit-vector
    :rtype: iterable[int]
    """
    out = np.zeros(length)
    out[indices] = 1
    return out

def remove_hot_overlap(input_vector, reference_vector):  # *
    """Removes overlap from input vector by setting reference vector hot indices to 0.

    :param input_vector: Vector to remove overlapping indices from
    :type input_vector: arraylike
    :param reference_vector: Vector to reference for overlapping indices
    :type reference_vector: arraylike
    :return: input_vector with overlapping indices set to 0
    :rtype: arraylike
    """
    input_vector = np.array(input_vector)
    reference_vector = np.array(reference_vector).astype(bool)
    input_vector[reference_vector] = 0
    return input_vector



from typing import Iterable


def type_check(arg_name:str, input, types:Iterable):
    if not any([isinstance(input, t) for t in types]):
        raise TypeError(f'Argument: "{arg_name}" should be (1) of types "{types}", not "{type(input)}"!')

def val_check(arg_name:str, input, allowed_vals:Iterable):
    if input not in allowed_vals:
        raise ValueError(f'Argument: "{arg_name}" should be (1) of "{allowed_vals}", not "{input}"!')




import numpy as np
from typing import Iterable, Union, Optional
from copy import copy

class Binarizer:
    active_operators = ['>', '<', '>=', '<=']
    
    __less_than = ['<', '≤']
    __greater_than = ['>', '≥']
    
    def __init__(self, values:Iterable, boundary:Union[float, int, np.number], active_operator:str,
                 qualifiers:Optional[Iterable]=None) -> None:
        self.__values = copy(values)
        self.__boundary = boundary
        self.__active_operator = active_operator
        self.__qualifiers = copy(qualifiers)
        
        if self.__active_operator not in self.active_operators:
            raise ValueError(f'Active operator must be one of the following: {self.active_operators}')
    
    @staticmethod
    def __equals_edge_case(relation:Optional[str], active_operator:str,
                           nan_greater_equal:bool=True) -> Union[int, float]:
        """Binarization handler for when value == threshold.

        Args:
            relation (str): Value qualifier. Set to one of: =, <, >, ≤, or ≥.
            active_operator (str): Condition to set values to active based on threshold. Set to one of: <, >, <=, or >=.
            nan_greater_equal (bool): Conditional to set any '≥' relation to np.nan. Defaults to True.

        Returns:
            Union[int, float]: 1 (active), 0 (inactive), or np.nan for '≥' special case.
        """
        if relation == '≥' and nan_greater_equal:  # Special case --> throw out bc ambiguous
            return np.nan
        return 1 if '=' in active_operator else 0
    
    @staticmethod
    def __active_operator_handler(is_greater_than:bool, active_operator:str) -> int:
        """Binarizes according to operator and (value > threshold) conditional.

        Args:
            is_greater_than (bool): Conditional for value > threshold.
            active_operator (str): Condition to set values to active based on threshold. Set to one of: <, >, <=, or >=.

        Returns:
            int: 1 (active) or 0 (inactive).
            
        Raises:
            ValueError: "active_operator" does not contain ">" or "<"
        """
        if '>' in active_operator:
            return 1 if is_greater_than else 0
        elif '<' in active_operator:
            return 0 if is_greater_than else 1
        else:
            raise(ValueError, '"active_operator" does not contain ">" or "<"')
    
    @staticmethod
    @np.vectorize
    def bin_qualifiers(value:Union[int, float, np.number], boundary:Union[int, float, np.number], relation:str,
                             active_operator:str) -> Union[int, float]:
        """Categorizes value as active (1), inactive (0), or conflict (np.nan) using qualifiers.

        Args:
            value (Union[int, float, np.number]): Number to evaluate threshold relationship.
            boundary (Union[int, float, np.number]): Binary decision boundary.
            relation (str): Value qualifier. Set to one of: =, <, >, ≤, or ≥.
            active_operator: Condition to set values to active based on threshold. Set to one of: <, >, <=, or >=.

        Returns:
            Union[int, float]: 1 (active), 0 (inactive), or np.nan for qualifier conflict / unrecognized relation.
        """
        
        greater_than_bound = None
        
        if np.isnan(value):
            return np.nan
        
        if value == boundary:
            return Binarizer.__equals_edge_case(relation, active_operator, nan_greater_equal=True)
        
        elif relation in Binarizer.__less_than:
            if value < boundary:
                greater_than_bound = False
            else:
                return np.nan  # Qualifier conflict
                
        elif relation in Binarizer.__greater_than:
            if value > boundary:
                greater_than_bound = True
            else:
                return np.nan  # Qualifier conflict

        elif relation == '=':
            greater_than_bound = True if value > boundary else False
            
        else:
            return np.nan  # Unrecognized relation
    
        return Binarizer.__active_operator_handler(greater_than_bound, active_operator)
    
    @staticmethod
    @np.vectorize
    def bin_no_qualifiers(value:Union[int, float, np.number], boundary:Union[int, float, np.number],
                                active_operator:str) -> Union[int, float]:
        """Categorizes value as active (1) or inactive (0).

        Args:
            value (Union[int, float, np.number]): Number to evaluate threshold relationship.
            boundary (Union[int, float, np.number]): Binary decision boundary.
            active_operator (str): Condition to set values to active based on threshold. Set to one of: <, >, <=, or >=.

        Returns:
            Union[int, float]: 1 (active), 0 (inactive), or np.nan special greater than equal edge case.
        """
        if value == boundary:
            return Binarizer.__equals_edge_case(None, active_operator, nan_greater_equal=True)
        else:
            greater_than_bound = (value > boundary)
            return Binarizer.__active_operator_handler(greater_than_bound, active_operator)
    
    def binarize(self) -> np.array:
        """Main binarize function.

        Returns:
            np.array: Binarized values.
        """
        if self.__qualifiers:
            return self.bin_qualifiers(value=self.__values, boundary=self.__boundary, relation=self.__qualifiers,
                                       active_operator=self.__active_operator)
        else:
            return self.bin_no_qualifiers(value=self.__values, boundary=self.__boundary,
                                          active_operator=self.__active_operator)



import unittest
import numpy as np

from stse import Binarizer


class TestBinarize(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        cls.test_values = [
            55,
            4,
            7,
            100,
            2000
        ]
        cls.qualifiers = [
            '<',
            '>',
            '<',
            '=',
            '='
        ]
        
        cls.boundary = 10
        return super().setUpClass()
    
    def test_bin_no_qualifiers(self):
        out = []
        for op in zip(Binarizer.active_operators):
            out.append(Binarizer.bin_no_qualifiers(self.test_values, self.boundary, op).tolist())
            
        self.assertEqual(
            out,
            [
                [tv > self.boundary for tv in self.test_values],
                [tv < self.boundary for tv in self.test_values],
                [tv >= self.boundary for tv in self.test_values],
                [tv <= self.boundary for tv in self.test_values]
            ]
        )
        
    def test_bin_qualifiers(self):
        out = []
        for op in Binarizer.active_operators:
            out.append(Binarizer.bin_qualifiers(self.test_values, self.boundary, self.qualifiers, op)
                       .tolist())
        
        expected = [
                [tv > self.boundary for tv in self.test_values],
                [tv < self.boundary for tv in self.test_values],
                [tv >= self.boundary for tv in self.test_values],
                [tv <= self.boundary for tv in self.test_values]
            ]
        
        expected = [2*[np.nan] + ex[2:] for ex in expected]
        
        self.assertTrue(
            np.allclose(
                out,
                expected,
                equal_nan=True
            )
        )
        
    def test_binarize(self):
        out = []
        for op in Binarizer.active_operators:
            binarizer = Binarizer(values=self.test_values, qualifiers=self.qualifiers, boundary=self.boundary,
                                  active_operator=op)
            out.append(binarizer.binarize().tolist())
            
        expected = [
                [tv > self.boundary for tv in self.test_values],
                [tv < self.boundary for tv in self.test_values],
                [tv >= self.boundary for tv in self.test_values],
                [tv <= self.boundary for tv in self.test_values]
            ]
        
        expected = [2*[np.nan] + ex[2:] for ex in expected]
        
        self.assertTrue(
            np.allclose(
                out,
                expected,
                equal_nan=True
            )
        )
            

if __name__ == '__main__':
    unittest.main()



from stse import bytes_vect
from stse import database
from stse import dataframes
from stse import dictionaries
from stse import duplicates
from stse import plotting
from stse import stats
from stse import styling
from stse import web
from stse import error_checking
from stse import dev_tools
from stse.Binarizer import Binarizer



from typing import Iterable
import numpy as np
from difflib import get_close_matches
import re
import mimetypes
from werkzeug.datastructures import FileStorage
import tempfile
import pandas as pd
import json
from typing import Union, List, Tuple


"""
    File name: dataframes.py
    Author: Jacob Gerlach
    Description: Assortment of basic pandas DataFrame cleaning operations.
    Notes:
        * = Function has associated unit test.
"""

def drop_val(df, col_name, value):  # *
    """Drops row from df where column entry is equal to value.

    Args:
        df (pandas DataFrame): DataFrame containing rows to drop.
        col_name (str): Name of column to investigate.
        value (pandas object): Value to drop instances of.

    Returns:
        pandas DataFrame: DataFrame with rows dropped.
    """
    return pull_not_val(df, col_name, value).reset_index(drop=True)

def pull_val(df, col_name, value):  # *
    """Retrieves rows from df where column entry is equal to value.

    Args:
        df (pandas DataFrame): DataFrame containing rows to pull.
        col_name (str): Name of column to investigate.
        value (pandas object): Value to pull instances of.

    Returns:
        pandas DataFrame: DataFrame where value is found.
    """
    return df.where(df[col_name] == value).dropna()

def pull_not_val(df, col_name, value):  # *
    """Retrieves rows from df where column entry is not equal to value.

    Args:
        df (pandas DataFrame): DataFrame containing rows to pull.
        col_name (str): Name of column to investigate.
        value (pandas object): Value to not pull instances of.

    Returns:
        pandas DataFrame: DataFrame where value is not found.
    """
    return df.where(df[col_name] != value).dropna()

def id_nearest_col(df, name, similarity_ratio=0.5, case_insensitive=True):  # *
    """Sends closest column to input column name provided that it is at least 50% similar. CASE INSENSITIVE.

    Args:
        df (pandas DataFrame): DataFrame containing header with columns of interest.
        name (str): Column name to search for.
        case_insensitive (bool, optional): If True, converts both to lower before comparing. Defaults to True.

    Returns:
        str: Most similar column name.
    """
    
    # Convert both to lowercase
    if case_insensitive:
        df_head = [col.lower() for col in df]
        name = name.lower()
    else:
        df_head = df.columns
    
    # Find closest match
    matches = get_close_matches(name, df_head, n=1, cutoff=similarity_ratio)
    if len(matches) <= 0:
        return None
    else:
        match_index = df_head.index(matches[0])
        return df.columns[match_index]
    
def remove_header_chars(df, chars, case_insensitive=True):  # *
    """Removes chars in string from df header.

    Args:
        df (pandas DataFrame): DataFrame containing header with chars to remove.
        chars (str): Continuous string of chars to individually remove.
        case_insensitive (bool, optional): If True, converts both to lower before comparing. Defaults to True.

    Returns:
        pandas DataFrame: DataFrame with header chars removed.
    """
    df = df.copy()
    
    # Convert both to lowercase
    if case_insensitive:
        chars += chars.upper() + chars.lower()
    
    for column_name in df.columns:
        translate_remove = column_name.maketrans('', '', chars)
        new_column_name = column_name.translate(translate_remove)
        df.rename(columns={column_name: new_column_name}, inplace=True)

    return df

def convert_to_nan(df, na=('', 'nan', 'none')):  # *
    """Converts all instances found in na to np.nan. Case insensitive.

    Args:
        df (pandas DataFrame): Input DataFrame
        na (iterable, optional): Contains patterns to convert. Defaults to ('', 'nan', 'none').

    Returns:
        pandas DataFrame: DataFrame replaced to nan.
    """
    na = [s.lower() for s in na]
    return df.applymap(lambda s: np.nan if str(s).lower() in na else s)

def non_num_to_nan(df, columns):  # *
    """Removes all non numeric (0-9) characters from columns specified.

    Args:
        df (pandas DataFrame): Contains non-numeric values.

    Returns:
        pandas DataFrame: Non-numerics replaced w np.nan.
    """
    
    # Cast to dataframe in case series (1 column)
    try:
        df = df[columns].applymap(lambda s: re.sub('[^0-9]', '', str(s))).replace('', np.nan)
    except AttributeError: 
        df[columns] = df[columns].map(lambda s: re.sub('[^0-9]', '', s)).replace('', np.nan)
        
    return df

def nan_col_indices(df, column_name):  # *
    # replace_blanks_w_na(df)
    return df[df[column_name].isnull()].index.tolist()

def remove_nan_cols(df):  # *
    """Removes columns that are ENTIRELY blank or NaN.

    Args:
        df (pandas DataFrame): DataFrame containing columns to drop.

    Returns:
        pandas DataFrame: DataFrame with columns dropped.
    """
    return convert_to_nan(df).dropna(axis=1, how="all")

def remove_nan_rows(df, value_column_names):  # *
    """Removes entire rows from df where value columns are blank or NaN.

    Args:
        df (pandas DataFrame): DataFrame containing column with NaN values.
        value_column_name (iterable): Contains column names to look for.

    Returns:
        pandas DataFrame: DataFrame with rows dropped where value is NaN.
    """
    return df.dropna(axis=0, subset=value_column_names).reset_index(drop=True)

def store_df(df, ext):
    
    buffer = tempfile.NamedTemporaryFile()
    if ext == 'csv':
        df.to_csv(buffer, index=False)
    elif ext == 'xlsx':
        df.to_excel(buffer, index=False)
    elif ext == 'tsv':
        df.to_csv(buffer, index=False, sep='\t')
    else:
        raise('Only "csv", "xlsx" and "tsv" extensions are accepted')
    
    type = mimetypes.guess_type(f'_.{ext}')[0]
    
    return FileStorage(
        stream=open(buffer.name, 'rb'),
        filename=f'test.{ext}',
        content_type=type,
    )
    
def df_2_json(df:pd.DataFrame) -> str:  # *
    """Converts from dataframe to JSON string.

    Args:
        df (pd.DataFrame): Data.

    Returns:
        str: JSON string data.
    """
    return json.dumps(df.to_numpy().tolist())

def z_norm(df:pd.DataFrame) -> pd.DataFrame:
    """Normalized dataframe to a mean of 0 and standard deviation of 1.

    Args:
        df (pd.DataFrame): Data to transform.

    Returns:
        pd.DataFrame: Normalized data.
    """
    return (df - df.mean())/df.std()

def compare(inner_df:pd.DataFrame, outer_df:pd.DataFrame, inner_column:str, outer_column:str,
            find_same:bool=True) -> pd.DataFrame:
    """Compares two DataFrames along two specified column names.

    Args:
        inner_df (pd.DataFrame): Data to search for in outer_df.
        outer_df (pd.DataFrame): Data to query with inner_df.
        inner_column (str): Name of column to compare along in inner_df.
        outer_column (str): Name of column to compare along in outer_df.
        find_same (bool, optional): If True returns same values, if False returns values in outer_df column not found in
            inner_df column. Defaults to True.

    Returns:
        pd.DataFrame: outer_df with condition specified by find_same.
    """
    outer_in_inner = outer_df[outer_column].isin(inner_df[inner_column])
    indices = outer_in_inner[(outer_in_inner if find_same else ~outer_in_inner)].index
    return outer_df.iloc[indices]

def concat(dfs:Iterable[pd.DataFrame], drop_uncommon_columns:bool=True) -> pd.DataFrame:
    """Wrapper for pd.concat. Stacks DataFrames and sorts.

    Args:
        dfs (Iterable[pd.DataFrame]): Collection of DataFrames.
        drop_uncommon_columns (bool, optional): Determines pd join method. Defaults to True.

    Returns:
        pd.DataFrame: Concatenated data.
    """
    return pd.concat(dfs, axis=0, ignore_index=True, join=('inner' if drop_uncommon_columns else 'outer'), sort=True)

def sync_na_drop(df:Union[pd.DataFrame, pd.Series], col_s:Union[str, List[str]], *iterables:Iterable,
                 all_na:bool=True) -> Tuple[pd.DataFrame, Union[list, List[list]]]:
    """Syncs NaN values in a dataframe with a list of iterables.

    Args:
        df (pd.DataFrame): Data to drop from.
        col_s (Union[str, list]): Name of single column or list of columns to subset for drop.
        all_na (bool): True, drops row only if all values in row are NaN. Else drops row if any values is NaN.
        Defaults to True.

    Returns:
        Tuple[List[Iterable], pd.DataFrame]: Tuple of list of masked iterables and NA dropped DataFrame.
    """
    if not all(len(i) == len(df) for i in iterables):
        raise ValueError('All iterables must be the same length as the dataframe.')
     
    f = df[col_s].isnull
    if isinstance(df[col_s], pd.DataFrame):
        is_na_mask = f().all(axis=1) if all_na else f().any(axis=1)
    elif isinstance(df[col_s], pd.Series):
        is_na_mask = f()
    
    if len(iterables) == 1:
        out = np.array(iterables[0])[~is_na_mask].tolist()
    else:
        out = [np.array(it)[~is_na_mask].tolist() for it in iterables]
        
    return (df.dropna(subset=list(col_s), how=('all' if all_na else 'any')),
            out)



import pandas as pd


"""
    File name: web.py
    Author: Jacob Gerlach
    Description: Assortment of basic web request operations.
    Notes:
        * = Function has associated unit test.
"""

def filemime_2_df(file, return_ext=False):  # *
    """Reads spreadsheet from werkzeug FileStorage object according to file mimetype.

    :param file: Werkzeug FileStorage input
    :type file: werkzeug FileStorage
    :param return_ext: If True the extension is returned with df as a tuple, defaults to False
    :type return_ext: bool, optional
    :return: pandas DataFrame if return_ext is False, else (pandas DataFrame, file extension)
    :rtype: pandas DataFrame or (pandas DataFrame, str)
    """

    # Get file MIME
    file_mime = file.content_type

    # Read CSV
    if file_mime == 'text/csv':
        file_ext = 'csv'

        # Read in data
        input_data = pd.read_csv(file)
        
    # Read TSV
    elif file_mime == 'text/tab-separated-values':
        file_ext = 'tsv'

        # Read in data
        input_data = pd.read_csv(file, sep='\t')
        
    # Read Excel
    elif (file_mime == 'application/vnd.ms-excel' or
        file_mime == 'application/vnd.openxmlformats-officedocument.spreadsheetml.sheet'):
        file_ext = 'xlsx'

        # Read in data
        input_data = pd.read_excel(file)

    # Unsupported filetype
    else:
        raise(f'"{file_mime}" is not a supported filetype.')


    return (input_data, file_ext) if return_ext else input_data

def filename_2_df(file):
    """Reads spreadsheet from werkzeug FileStorage object according to filename string extension.

    :param file: Werkzeug FileStorage input
    :type file: werkzeug FileStorage
    :return: spreadsheet read from file object
    :rtype: pandas DataFrame
    """
    
    # Split extension from string
    ext = file.filename.split('.')[-1]
    
    # Map extension to pandas function
    read_function = {
        'csv': pd.read_csv,
        'xlsx': pd.read_excel,
        'xls': pd.read_excel,
        'tsv': lambda x: pd.read_csv(x, sep='\t')
    }
    
    try:
        return read_function[ext](file)
    except Exception as e:  # Unsupported filetype
        print(e)



import unittest
import pandas as pd

from stse import duplicates


class TestDuplicates(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        cls.test_df = pd.DataFrame({'aa': ['5', '5', '6', 5, 6, '6'], 'b': [1, 2, 3, 4, 5, 6], 
                                     'c': ['a', 'a', 'b', 'b', 'b', 'b']})
        cls.filt_not_dup_sol = pd.DataFrame({'aa': ['5', '6', 5, 6], 'b': [1, 3, 4, 5], 'c': ['a', 'b', 'b', 'b']}, 
                                             index=[0, 2, 3, 4])
        cls.filt_dup_sol = pd.DataFrame({'aa': ['5', '6'], 'b': [2, 6], 'c': ['a', 'b']}, index=[1, 5])
        cls.filt_multi_not_dup_sol = pd.DataFrame({'aa': ['5', '6'], 'b': [1, 3], 'c': ['a', 'b']}, 
                                                   index=[0, 2])
        cls.filt_multi_dup_sol = pd.DataFrame({'aa': ['5', 5, 6, '6'], 'b': [2, 4, 5, 6], 
                                                'c': ['a', 'b', 'b', 'b']}, index=[1, 3, 4, 5])
        cls.average_sol = pd.DataFrame({'aa': ['5', '6', 5, 6], 'b': [1.5, 4.5, 4, 5], 'c': ['a', 'b', 'b', 'b']}, 
                                        index=[0, 2, 3, 4])
        return super().setUpClass()
    
    def test_filter(self):
        # Test filtering out duplicates
        assert duplicates.filter(self.test_df, ['aa']).equals(self.filt_not_dup_sol)
        
        # Test filtering out everything EXCEPT duplicates
        assert duplicates.filter(self.test_df, ['aa'], return_duplicates=True).equals(self.filt_dup_sol)
        
        # Test filtering across multiple subsets
        assert duplicates.filter(self.test_df, ['aa', 'c'], how='any').equals(self.filt_multi_not_dup_sol)
        assert duplicates.filter(self.test_df, ['aa', 'c'], how='any', return_duplicates=True).equals(
            self.filt_multi_dup_sol)
    
    def test_indices(self):
        self.assertEqual(
            duplicates.indices(self.test_df, 'c'), 
            [1, 3, 4, 5]
        )
        
    def test_remove(self):
        assert duplicates.remove(self.test_df, ['aa']).equals(self.filt_not_dup_sol)
        
    def test_average(self):
        assert duplicates.average(self.test_df, ['aa'], average_by='b').equals(self.average_sol)


if __name__ == '__main__':
    unittest.main()



import pandas as pd
import numpy as np


"""
    File name: duplicates.py
    Author: Jacob Gerlach
    Description: Assortment of pandas DataFrame operations for handling duplicates.
    Notes:
        * = Function has associated unit test.
"""

def filter(df, subsets, return_duplicates=False, how='all'):  # *
    """Filters duplicates to either return DataFrame subset of duplicates or non-duplicates.

    Args:
        df (pandas DataFrame): DataFrame to perform duplicate search on.
        subsets (list): List of column names to compare in duplicate search.
                        All must match to be considered a duplicate.
        return_duplicates (bool, optional): If True, returns non-duplicate df else returns duplicate df.
                                            Defaults to False.
        how (str): 

    Returns:
        pandas DataFrame: Filtered df.
    """
    if how == 'any' and len(subsets) > 1:
        # Checks for occurances in ANY column of 'subsets'
        index_arr = np.array([])
        for col in subsets:     
            index_arr = np.concatenate((index_arr, indices(df, col)))
        
        index_arr = np.unique(index_arr)
        
        return df.loc[df.index.isin(index_arr)] if return_duplicates else df.loc[~df.index.isin(index_arr)]
    
    elif how == 'all':
        # Duplicates in ALL columns of 'subsets' must agree
        duplicates = pd.DataFrame.duplicated(df, subset=subsets)
        
    else:
        # Raise error of how is not 'any' or 'all'
        raise ValueError('how argument must be either \'any\' or \'all\'')
    
    return df[duplicates] if return_duplicates else df[~duplicates]

def remove(df, subsets):  # *
    """Removes duplicates from df. Keeps first occurance in df.

    Args:
        df (pandas DataFrame): DataFrame to perform dupicate search on.
        subsets (list): List of column names to compare in duplicate search.
                        All must match to be considered a duplicate.

    Returns:
        pandas DataFrame: DataFrame with duplicates removed.
    """
    return filter(df, subsets=subsets)

def indices(df, column):  # *
    """Finds indices of rows in df column that hold duplicate values.

    :param df: Dataframe to perform duplicate search on
    :type df: pandas DataFrame
    :param column: Name of column in df
    :type column: str
    :return: Indices
    :rtype: list
    """
    return list(df[pd.DataFrame.duplicated(df, subset=[column])].index)

def average(df, subsets, average_by):  # *
    """Averages duplicates from df.

    Args:
        df (pandas DataFrame): DataFrame to perform duplicate search on.
        subsets (list): List of column names to compare in duplicate search. 
                        All must match to be considered a duplicate.
        average_by (str): Column name to average along.

    Returns:
        pandas DataFrame: DataFrame with duplicates averaged.
    """
    df_copy = df.copy()
    
    # Filter out duplicates for merge df
    out = filter(df_copy, subsets)
    orig_index = out.index
        
    # Ensure average_by exists in the dataframe
    if average_by not in df.columns:
        raise ValueError('average_by column not found in DataFrame')
    
    # Cast value column to float (for avg computation)
    try:
        df_copy[average_by] = df_copy[average_by].astype(float)
    except ValueError:
        raise ValueError('average_by column does not contain float castable values')
    
    # Group duplicate rows and aggregate by mean (adds mean to end of value column)
    avg_value = df_copy.groupby(by=subsets).agg({average_by: ['mean']})
    avg_value.columns = [col[0] for col in avg_value.columns.values]

    # Drop original value column ('value' --> 'valuemean')
    out = out.drop(columns=[average_by])  # drop original values

    # Add averaged values to dataframe and preserve original index
    out = pd.merge(out, avg_value, on=subsets).set_index(orig_index)
    out = out[df_copy.columns]
    
    return out



import plotly.express as px


class ConfigHeatmap:
    def __init__(self, array, options):
        self.array = array
        self.options = options

        self.fig = px.imshow(array, color_continuous_scale=options['color'])

    def title(self):
        self.fig.update_layout(title_text=self.options['title'])

        if self.options['center_title']:
            self.fig.update_layout(title_x=0.5)

        if self.options['title_pos'] == 'bottom':
            self.fig.update_layout(title_y=0.1)

    def axes(self):
        self.fig.update_xaxes(visible=self.options['x_axis'])
        self.fig.update_yaxes(visible=self.options['y_axis'])

        self.fig.update_layout(xaxis_side=self.options['x_axis_pos'], yaxis_side=self.options['y_axis_pos'])

    def config_all(self):
        self.title()
        self.axes()



import unittest
import pandas as pd

from stse import web, dataframes


class TestWeb(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        cls.df = pd.DataFrame({
            'a': [1, 2, 3, 4],
            'b': ['a', 'b', 'c', 'd']
        })
        return super().setUpClass()
    
    def test_filemime_2_df(self):
        # Test csv
        mock_file = dataframes.store_df(self.df, 'csv')
        pd.testing.assert_frame_equal(
            web.filemime_2_df(mock_file), 
            self.df
        )
        mock_file.close()
        
        # Test excel
        mock_file = dataframes.store_df(self.df, 'xlsx')
        pd.testing.assert_frame_equal(
            web.filemime_2_df(mock_file), 
            self.df
        )
        mock_file.close()
        
        # Test csv
        mock_file = dataframes.store_df(self.df, 'tsv')
        pd.testing.assert_frame_equal(
            web.filemime_2_df(mock_file), 
            self.df
        )
        mock_file.close()
        
    def test_filename_2_df(self):
        # Test csv
        mock_file = dataframes.store_df(self.df, 'csv')
        pd.testing.assert_frame_equal(
            web.filemime_2_df(mock_file), 
            self.df
        )
        mock_file.close()
        
        # Test excel
        mock_file = dataframes.store_df(self.df, 'xlsx')
        pd.testing.assert_frame_equal(
            web.filemime_2_df(mock_file), 
            self.df
        )
        mock_file.close()
        
        # Test csv
        mock_file = dataframes.store_df(self.df, 'tsv')
        pd.testing.assert_frame_equal(
            web.filemime_2_df(mock_file), 
            self.df
        )
        mock_file.close()


if __name__ == '__main__':
    unittest.main()



import unittest
import pandas as pd
import numpy as np
import json

from stse import dataframes


class TestDataframes(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        cls.test_df = pd.DataFrame({'a': [4, '5', '6'], 'b': ['6', '5', '4']})
        cls.mol_test_df = pd.DataFrame({'SMILES': [4, '5', '6'], 'Molecule': ['6', '5', '4'], 'InChi': [5, 6, 8]})
        cls.nan_test_df = pd.DataFrame({'a': ['', '5', '6'], 'b': ['a', '8', 'None']})
        cls.blank_solution = pd.DataFrame({'a': [np.nan, '5', '6'], 'b': ['a', '8', np.nan]})
        cls.non_num_solution = pd.DataFrame({'a': [np.nan, '5', '6'], 'b': [np.nan, '8', np.nan]})
        return super().setUpClass()
    
    def test_drop_val(self):
        # Test number
        dropped1 = dataframes.drop_val(self.test_df, 'a', 4)
        assert dropped1.equals(pd.DataFrame({'a': ['5', '6'], 'b': ['5', '4']}))
        
        # Test string
        dropped2 = dataframes.drop_val(self.test_df, 'a', '4')
        assert dropped2.equals(self.test_df)
        
        # Test b column
        dropped3 = dataframes.drop_val(self.test_df, 'b', '4')
        assert dropped3.equals(pd.DataFrame({'a': [4, '5'], 'b': ['6', '5']}))
        
    def test_pull_val(self):
        # Test number
        pulled1 = dataframes.pull_val(self.test_df, 'a', 4)
        assert pulled1.equals(pd.DataFrame({'a': [4], 'b': ['6']}).astype(object))
        
        # Test string
        pulled2 = dataframes.pull_val(self.test_df, 'a', '6')
        assert pulled2.equals(pd.DataFrame({'a': ['6'], 'b': ['4']}, index=[2]))
        
        # Test b column
        pulled3 = dataframes.pull_val(self.test_df, 'b', '5')
        assert pulled3.equals(pd.DataFrame({'a': ['5'], 'b': ['5']}, index=[1]))
        
    def test_pull_not_val(self):
        # Test number
        pulled1 = dataframes.pull_not_val(self.test_df, 'a', 4)
        assert pulled1.equals(pd.DataFrame({'a': ['5', '6'], 'b': ['5', '4']}, index=[1, 2]).astype(object))
        
        # Test string
        pulled2 = dataframes.pull_not_val(self.test_df, 'a', '6')
        assert pulled2.equals(pd.DataFrame({'a': [4, '5'], 'b': ['6', '5']}))
        
        # Test b column
        pulled3 = dataframes.pull_not_val(self.test_df, 'b', '5')
        assert pulled3.equals(pd.DataFrame({'a': [4, '6'], 'b': ['6', '4']}, index=[0, 2]))
        
        # Test pull invalid number --> returns what was input
        pulled4 = dataframes.pull_not_val(self.test_df, 'b', '3')
        assert pulled4.equals(self.test_df)
        
    def test_id_nearest_col(self):
        # Test similar variations
        assert dataframes.id_nearest_col(self.mol_test_df, 'smiles') == 'SMILES'
        assert dataframes.id_nearest_col(self.mol_test_df, 'smIles') == 'SMILES'
        assert dataframes.id_nearest_col(self.mol_test_df, 'smes') == 'SMILES'
        
        # Test null return
        assert dataframes.id_nearest_col(self.mol_test_df, 's') == None
        
    def test_remove_header_chars(self):
        test_df1 = self.mol_test_df.copy()
        test_df2 = self.mol_test_df.copy()
        test_df3 = self.mol_test_df.copy()
        test_df1.columns = ['MILE', 'Molecule', 'InChi']
        test_df2.columns = ['SMLES', 'Molecule', 'nCh']
        test_df3.columns = ['SMILES', 'Molecule', 'InCh']
        
        # Test w/ case sensitive on
        assert dataframes.remove_header_chars(self.mol_test_df, 's').equals(test_df1)
        assert dataframes.remove_header_chars(self.mol_test_df, 'S').equals(test_df1)
        
        # Test w/ case sensitive off
        assert dataframes.remove_header_chars(self.mol_test_df, 's', case_insensitive=False).equals(self.mol_test_df)
        assert dataframes.remove_header_chars(self.mol_test_df, 'S', case_insensitive=False).equals(test_df1)
        
        # Test multiple instances
        assert dataframes.remove_header_chars(self.mol_test_df, 'i').equals(test_df2)
        assert dataframes.remove_header_chars(self.mol_test_df, 'i', case_insensitive=False).equals(test_df3)
        
    def test_convert_to_nan(self):
        # Test replace nan
        na_df = dataframes.convert_to_nan(self.nan_test_df)
        assert na_df.equals(self.blank_solution)
        
        # Test case sensitivity
        na_df = dataframes.convert_to_nan(self.nan_test_df, na=('A'))
        assert na_df.equals(pd.DataFrame({'a': ['', '5', '6'], 'b': [np.nan, '8', 'None']}))
        
    def test_non_num_to_nan(self):
        # Test on only first column
        na_df = dataframes.non_num_to_nan(self.nan_test_df, self.nan_test_df.columns[0])
        assert na_df.equals(pd.DataFrame({'a': [np.nan, '5', '6'], 'b': ['a', '8', 'None']}))
        
        # Test on only second column (unchanged)
        na_df = dataframes.non_num_to_nan(self.nan_test_df, self.nan_test_df.columns[1])
        assert na_df.equals(self.nan_test_df)
        
        # Test on both columns
        na_df = dataframes.non_num_to_nan(self.nan_test_df, self.nan_test_df.columns)
        assert na_df.equals(self.non_num_solution)
        
    def test_nan_col_indices(self):
        # Test each col separately
        assert dataframes.nan_col_indices(self.blank_solution, self.blank_solution.columns[0]) == [0]
        assert dataframes.nan_col_indices(self.blank_solution, self.blank_solution.columns[1]) == [2]
        
    def test_remove_nan_cols(self):
        nan_col_df = self.test_df.copy()
        
        # Append new columns
        nan_col_df['c'] = 3*['None']
        nan_col_df['d'] = 3*['nan']
        nan_col_df['e'] = 3*['']
        nan_col_df['f'] = 3*[np.nan]
        
        # Check new columns were added successfully
        assert len(nan_col_df.columns) == 6
        
        # Test blank column removal
        assert dataframes.remove_nan_cols(nan_col_df).equals(self.test_df)
        
    def test_remove_nan_rows(self):
        self.non_num_solution = pd.DataFrame({'a': [np.nan, '5', '6'], 'b': [np.nan, '8', np.nan]})
        
        col_a = dataframes.remove_nan_rows(self.blank_solution, ['a'])
        assert col_a.equals(pd.DataFrame({'a': ['5', '6'], 'b': ['8', np.nan]}))
        
        col_b = dataframes.remove_nan_rows(self.blank_solution, ['b'])
        assert col_b.equals(pd.DataFrame({'a': [np.nan, '5'], 'b': ['a', '8']}))
        
        both_cols = dataframes.remove_nan_rows(self.blank_solution, self.blank_solution.columns)
        assert both_cols.equals(pd.DataFrame({'a': ['5'], 'b': ['8']}))
        
    def test_df_2_json(self):
        out = dataframes.df_2_json(self.test_df)
        expected = '[[4, "6"], ["5", "5"], ["6", "4"]]'
        
        self.assertEqual(
            out,
            expected
        )
        
        # Ensure valid json
        try:
            json.loads(out)
        except Exception as e:
            self.fail(e)
            
    def test_z_norm(self):
        test_df = pd.DataFrame({
            'a': [1, 4, 7],
            'b': [5, 9, 10]
        })
        z_norm_df = dataframes.z_norm(test_df)
        
        # Zero mean
        self.assertFalse(
            np.allclose(
                test_df.mean(),
                np.array(2*[0], dtype=float)
            )
        )
        self.assertTrue(
            np.allclose(
                z_norm_df.mean(),
                np.zeros(2)
            )
        )
        
        # Standard dev. of 1
        self.assertFalse(
            np.allclose(
                test_df.std(),
                np.ones(2)
            )
        )
        self.assertTrue(
            np.allclose(
                z_norm_df.std(),
                np.ones(2)
            )
        )
        
    def test_compare(self):
        inner = pd.DataFrame({
            'a': [1, 2, 3, 4],
            'b': [2, 3, 5, 6]
        })
        outer = pd.DataFrame({
            'c': [2, 3, 5, 6, 7, 8, 9, 10]
        })
        
        # Find same
        expected = outer.iloc[:2]
        out = dataframes.compare(inner, outer, 'a', 'c', find_same=True)
        self.assertTrue(
            out.equals(expected)
        )
        
        # Find unique
        expected = outer.iloc[2:]
        out = dataframes.compare(inner, outer, 'a', 'c', find_same=False)
        self.assertTrue(
            out.equals(expected)
        )
    
    def test_concat(self):
        df1 = pd.DataFrame({
            'a': [1, 2, 3],
            'c': [4, 5, 6],
            'b': [7, 8, 9]
        })
        df2 = pd.DataFrame({
            'b': [10, 11],
            'c': [13, 14],
            'd': [16, 17]
        })
        
        # Drop uncommon
        expected1 = pd.DataFrame({
            'b': [7, 8, 9, 10, 11],
            'c': [4, 5, 6, 13, 14]
        })
        out1 = dataframes.concat([df1, df2], drop_uncommon_columns=True)
        self.assertTrue(
            out1.equals(expected1)
        )
        
        # Keep uncommon, replace blank w/ nan
        expected2 = pd.DataFrame({
            'a': [1, 2, 3, np.nan, np.nan],
            'b': [7, 8, 9, 10, 11],
            'c': [4, 5, 6, 13, 14],
            'd': [np.nan, np.nan, np.nan, 16, 17]
        })
        out2 = dataframes.concat([df1, df2], drop_uncommon_columns=False)
        self.assertTrue(
            out2.equals(expected2)
        )
        
        # Dropna is possible
        self.assertTrue(
            out2.dropna(axis=1).equals(expected1)
        )
        
    def test_sync_na_drop(self):
        df1 = pd.DataFrame({
            'a': [1, np.nan, 3],
            'b': [4, np.nan, np.nan],
            'c': [7, 8, 9]
        })
        iter1 = ['a', 'b', 'c']
        
        # One column
        out_df1, out_iter1 = dataframes.sync_na_drop(df1, 'a', iter1, all_na=True)
        out_df2, out_iter2 = dataframes.sync_na_drop(df1, 'a', iter1, all_na=False)
        
        self.assertTrue(
            out_df1.equals(
                df1.dropna(subset=['a'])
            )
        )
        self.assertTrue(  # Same bc only one column
            out_df1.equals(
                out_df2
            )
        )
        self.assertEqual(
            out_iter1,
            ['a', 'c'],
            out_iter2
        )
        
        # Two columns, all_na=True
        out_df3, out_iter3 = dataframes.sync_na_drop(df1, ['a', 'c'], iter1, all_na=True)
        
        self.assertTrue(  # No change
            out_df3.equals(
                df1
            )
        )
        self.assertEqual(  # No change
            out_iter3,
            iter1
        )
        
        # Two columns, all_na=False
        out_df4, out_iter4 = dataframes.sync_na_drop(df1, ['a', 'c'], iter1, all_na=False)
        
        self.assertTrue(
            out_df4.equals(
                df1.dropna(subset=['a'])
            )
        )
        self.assertEqual(
            out_iter4,
            ['a', 'c']
        )
        
        # 'a' and 'b' columns, all_na=True
        out_df5, out_iter5 = dataframes.sync_na_drop(df1, ['a', 'b'], iter1, all_na=True)
        
        self.assertTrue(
            out_df5.equals(
                df1.dropna(subset=['a', 'b'], how='all')
            )
        )
        self.assertEqual(
            out_iter5,
            ['a', 'c']
        )
        
        # a' and 'b' columns, all_na=False
        out_df6, out_iter6 = dataframes.sync_na_drop(df1, ['a', 'b'], iter1, all_na=False)
        
        self.assertTrue(
            out_df6.equals(
                df1.dropna(subset=['a', 'b'], how='any')
            )
        )
        self.assertEqual(
            out_iter6,
            ['a']
        )
        
        # Multiple iterables
        iter2 = ['d', 'e', 'f']
        _, out_iter7 = dataframes.sync_na_drop(df1, ['a', 'b'], iter1, iter2, all_na=False)
        
        self.assertEqual(
            out_iter7,
            [
                ['a'],
                ['d']
            ]
        )
        
        # Raise length error
        iter3 = 5*['h']
        with self.assertRaises(ValueError):
            dataframes.sync_na_drop(df1, ['a', 'b'], iter1, iter3, all_na=True)
        

if __name__ == '__main__':
    unittest.main()
